from unittest import IsolatedAsyncioTestCase

from ...admin.base_server import BaseAdminServer
from ...askar.profile import AskarProfileManager
from ...config.base_context import ContextBuilder
from ...config.injection_context import InjectionContext
from ...connections.base_manager import BaseConnectionManager
from ...connections.models.conn_record import ConnRecord
from ...connections.models.connection_target import ConnectionTarget
from ...connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from ...core.event_bus import EventBus, MockEventBus
from ...core.profile import ProfileManager
from ...core.protocol_registry import ProtocolRegistry
from ...ledger.base import BaseLedger
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...protocols.coordinate_mediation.mediation_invite_store import MediationInviteRecord
from ...protocols.coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ...protocols.out_of_band.v1_0.models.oob_record import OobRecord
from ...resolver.did_resolver import DIDResolver
from ...storage.base import BaseStorage
from ...storage.error import StorageNotFoundError
from ...storage.record import StorageRecord
from ...storage.type import RECORD_TYPE_ACAPY_STORAGE_TYPE
from ...tests import mock
from ...transport.inbound.manager import InboundTransportManager
from ...transport.inbound.message import InboundMessage
from ...transport.inbound.receipt import MessageReceipt
from ...transport.outbound.base import OutboundDeliveryError, QueuedOutboundMessage
from ...transport.outbound.message import OutboundMessage
from ...transport.outbound.status import OutboundSendStatus
from ...transport.pack_format import PackWireFormat
from ...transport.wire_format import BaseWireFormat
from ...utils.stats import Collector
from ...utils.testing import create_test_profile
from ...version import RECORD_TYPE_ACAPY_VERSION
from ...wallet.did_info import DIDInfo
from ...wallet.did_method import SOV, DIDMethods
from ...wallet.key_type import ED25519, KeyTypes
from .. import conductor as test_module


class Config:
    test_settings = {
        "admin.webhook_urls": ["http://sample.webhook.ca"],
        "wallet.type": "askar",
        "wallet.key": "insecure",
    }
    test_settings_admin = {
        "admin.webhook_urls": ["http://sample.webhook.ca"],
        "admin.enabled": True,
        "wallet.type": "askar",
        "wallet.key": "insecure",
        "auto_provision": True,
    }
    test_settings_with_queue = {"queue.enable_undelivered_queue": True}


class TestDIDs:
    test_seed = "testseed000000000000000000000001"
    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_endpoint = "http://localhost"

    test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
    test_target_verkey = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"

    def make_did_doc(self, did, verkey):
        doc = DIDDoc(did=did)
        controller = did
        ident = "1"
        pk_value = verkey
        pk = PublicKey(
            did, ident, pk_value, PublicKeyType.ED25519_SIG_2018, controller, False
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            did, "indy", "IndyAgent", recip_keys, router_keys, self.test_endpoint
        )
        doc.set(service)
        return doc, pk


class StubContextBuilder(ContextBuilder):
    def __init__(self, settings):
        super().__init__(settings)
        self.wire_format = mock.create_autospec(PackWireFormat())

    async def build_context(self) -> InjectionContext:
        context = InjectionContext(settings=self.settings, enforce_typing=False)
        context.injector.bind_instance(ProfileManager, AskarProfileManager())
        context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())
        context.injector.bind_instance(BaseWireFormat, self.wire_format)
        context.injector.bind_instance(DIDMethods, DIDMethods())
        context.injector.bind_instance(DIDResolver, DIDResolver([]))
        context.injector.bind_instance(EventBus, MockEventBus())
        context.injector.bind_instance(KeyTypes, KeyTypes())
        context.injector.bind_instance(
            InboundTransportManager,
            mock.MagicMock(InboundTransportManager, autospec=True),
        )
        context.injector.bind_instance(BaseConnectionManager, mock.MagicMock())
        return context


class StubCollectorContextBuilder(StubContextBuilder):
    async def build_context(self) -> InjectionContext:
        context = await super().build_context()
        context.injector.bind_instance(Collector, Collector())
        return context


class TestConductor(IsolatedAsyncioTestCase, Config, TestDIDs):
    async def test_startup_version_record_exists(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(
                    return_value=["v0.7.4", "0.7.5", "v0.8.0-rc1", "v8.0.0", "v0.8.1-rc2"]
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module, "LoggingConfigurator", autospec=True
            ) as mock_logger,
        ):
            await conductor.setup()

            mock_inbound_mgr.return_value.setup.assert_awaited_once()
            mock_outbound_mgr.return_value.setup.assert_awaited_once()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}

            async with test_profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(
                    StorageRecord(RECORD_TYPE_ACAPY_VERSION, "v0.7.3")
                )

                await conductor.start()

                mock_inbound_mgr.return_value.start.assert_awaited_once_with()
                mock_outbound_mgr.return_value.start.assert_awaited_once_with()

                mock_logger.print_banner.assert_called_once()

                upgrade_record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_VERSION, tag_query={}
                )

                assert upgrade_record != "v0.7.3"

                await conductor.stop()

                mock_inbound_mgr.return_value.stop.assert_awaited_once_with()
                mock_outbound_mgr.return_value.stop.assert_awaited_once_with()

    async def test_startup_version_no_upgrade_add_record(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(
                    return_value=["v0.7.4", "0.7.5", "v0.8.0-rc1", "v8.0.0", "v0.8.1-rc2"]
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}
            await conductor.start()
            await conductor.stop()

    async def test_startup_version_force_upgrade(self):
        test_settings = {
            "admin.webhook_urls": ["http://sample.webhook.ca"],
            "upgrade.from_version": "v0.7.5",
            "upgrade.force_upgrade": True,
            "wallet.type": "askar",
            "wallet.key": "insecure",
        }
        builder: ContextBuilder = StubContextBuilder(test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(
                    return_value=["v0.7.4", "0.7.5", "v0.8.0-rc1", "v8.0.0", "v0.8.1-rc2"]
                ),
            ),
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }

            async with test_profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(
                    StorageRecord(RECORD_TYPE_ACAPY_VERSION, "v0.7.3")
                )

                await conductor.setup()
                mock_inbound_mgr.return_value.registered_transports = {}
                mock_outbound_mgr.return_value.registered_transports = {}
                await conductor.start()
                await conductor.stop()

        test_profile = await create_test_profile(None, await builder.build_context())
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module, "LoggingConfigurator", autospec=True
            ) as mock_logger,
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(return_value=["v0.8.0-rc1", "v8.0.0", "v0.8.1-rc1"]),
            ),
            mock.patch.object(test_module, "ledger_config"),
            mock.patch.object(test_module.Conductor, "check_for_valid_wallet_type"),
        ):
            await conductor.setup()
            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}
            await conductor.start()
            mock_logger.print_banner.assert_called_once()
            await conductor.stop()

    async def test_startup_version_record_not_exists(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module, "LoggingConfigurator", autospec=True
            ) as mock_logger,
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
            mock.patch.object(
                BaseStorage,
                "find_record",
                mock.CoroutineMock(
                    side_effect=[mock.MagicMock(value="askar"), StorageNotFoundError()]
                ),
            ),
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(return_value=["v0.8.0-rc1", "v8.0.0", "v0.8.1-rc1"]),
            ),
            mock.patch.object(
                test_module,
                "upgrade",
                mock.CoroutineMock(),
            ),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            mock_inbound_mgr.return_value.setup.assert_awaited_once()
            mock_outbound_mgr.return_value.setup.assert_awaited_once()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}

            await conductor.start()

            mock_inbound_mgr.return_value.start.assert_awaited_once_with()
            mock_outbound_mgr.return_value.start.assert_awaited_once_with()

            mock_logger.print_banner.assert_called_once()

            await conductor.stop()

            mock_inbound_mgr.return_value.stop.assert_awaited_once_with()
            mock_outbound_mgr.return_value.stop.assert_awaited_once_with()

    async def test_startup_admin_server_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(test_module, "LoggingConfigurator", autospec=True),
            mock.patch.object(
                test_module, "AdminServer", mock.MagicMock()
            ) as mock_admin_server,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            mock_admin_server.side_effect = ValueError()
            with self.assertRaises(ValueError):
                await conductor.setup()

    async def test_startup_no_public_did(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module, "LoggingConfigurator", autospec=True
            ) as mock_logger,
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
        ):
            mock_outbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.enqueue_message = mock.CoroutineMock()
            mock_outbound_mgr.return_value.start = mock.CoroutineMock()
            mock_outbound_mgr.return_value.stop = mock.CoroutineMock()
            await conductor.setup()

            mock_inbound_mgr.return_value.setup.assert_awaited_once()
            mock_outbound_mgr.return_value.setup.assert_awaited_once()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}

            # Doesn't raise
            await conductor.start()

            mock_inbound_mgr.return_value.start.assert_awaited_once_with()
            mock_outbound_mgr.return_value.start.assert_awaited_once_with()

            mock_logger.print_banner.assert_called_once()

            await conductor.stop()

            mock_inbound_mgr.return_value.stop.assert_awaited_once_with()
            mock_outbound_mgr.return_value.stop.assert_awaited_once_with()

    async def test_stats(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
        ):
            mock_inbound_mgr.return_value.sessions = ["dummy"]
            mock_outbound_mgr.return_value.outbound_buffer = [
                mock.MagicMock(state=QueuedOutboundMessage.STATE_ENCODE),
                mock.MagicMock(state=QueuedOutboundMessage.STATE_DELIVER),
            ]
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }

            await conductor.setup()

            stats = await conductor.get_stats()
            assert all(
                x in stats
                for x in [
                    "in_sessions",
                    "out_encode",
                    "out_deliver",
                    "task_active",
                    "task_done",
                    "task_failed",
                    "task_pending",
                ]
            )

    async def test_inbound_message_handler(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with mock.patch.object(
            conductor.dispatcher, "queue_message", autospec=True
        ) as mock_dispatch_q:
            message_body = "{}"
            receipt = MessageReceipt(direct_response_mode="snail mail")
            message = InboundMessage(message_body, receipt)

            conductor.inbound_message_router(
                conductor.context, message, can_respond=False
            )

            mock_dispatch_q.assert_called_once()
            assert mock_dispatch_q.call_args[0][0] is conductor.context
            assert mock_dispatch_q.call_args[0][1] is message
            assert mock_dispatch_q.call_args[0][2] == conductor.outbound_message_router
            assert callable(mock_dispatch_q.call_args[0][3])

    async def test_inbound_message_handler_ledger_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with (
            mock.patch.object(
                conductor.dispatcher, "queue_message", autospec=True
            ) as mock_dispatch_q,
            mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger,
        ):
            mock_dispatch_q.side_effect = test_module.LedgerConfigError("ledger down")

            message_body = "{}"
            receipt = MessageReceipt(direct_response_mode="snail mail")
            message = InboundMessage(message_body, receipt)

            with self.assertRaises(test_module.LedgerConfigError):
                conductor.inbound_message_router(
                    conductor.context, message, can_respond=False
                )

            mock_dispatch_q.assert_called_once()
            mock_logger.error.assert_called_once()

    async def test_outbound_message_handler_return_route(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)
        test_to_verkey = "test-to-verkey"
        test_from_verkey = "test-from-verkey"
        test_profile = await create_test_profile(None, await builder.build_context())

        with mock.patch.object(
            test_module,
            "wallet_config",
            return_value=(
                test_profile,
                DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
            ),
        ):
            await conductor.setup()

        bus = conductor.root_profile.inject(EventBus)

        payload = "{}"
        message = OutboundMessage(payload=payload)
        message.reply_to_verkey = test_to_verkey
        receipt = MessageReceipt()
        receipt.recipient_verkey = test_from_verkey

        with mock.patch.object(
            conductor.inbound_transport_manager, "return_to_session"
        ) as mock_return:
            mock_return.return_value = True

            status = await conductor.outbound_message_router(
                conductor.root_profile, message
            )
            assert status == OutboundSendStatus.SENT_TO_SESSION
            assert bus.events
            assert bus.events[0][1].topic == status.topic
            assert bus.events[0][1].payload == message
            mock_return.assert_called_once_with(message)

    async def test_outbound_message_handler_with_target(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            assert conductor.root_profile
            bus = conductor.root_profile.inject(EventBus)

            payload = "{}"
            target = ConnectionTarget(
                endpoint="endpoint", recipient_keys=(), routing_keys=(), sender_key=""
            )
            message = OutboundMessage(payload=payload, target=target)

            status = await conductor.outbound_message_router(
                conductor.root_profile, message
            )
            assert status == OutboundSendStatus.QUEUED_FOR_DELIVERY
            assert bus.events
            assert bus.events[0][1].topic == status.topic
            assert bus.events[0][1].payload == message
            mock_outbound_mgr.return_value.enqueue_message.assert_called_once_with(
                conductor.root_profile, message
            )

    async def test_outbound_message_handler_with_connection(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            conn_mgr = mock.MagicMock(autospec=True)
            conn_mgr.get_connection_targets = mock.CoroutineMock()
            conductor.context.injector.bind_instance(BaseConnectionManager, conn_mgr)

            assert conductor.root_profile
            bus = conductor.root_profile.inject(EventBus)

            payload = "{}"
            connection_id = "connection_id"
            message = OutboundMessage(payload=payload, connection_id=connection_id)

            status = await conductor.outbound_message_router(
                conductor.root_profile, message
            )

            assert status == OutboundSendStatus.QUEUED_FOR_DELIVERY
            assert bus.events
            assert bus.events[0][1].topic == status.topic
            assert bus.events[0][1].payload == message

            conn_mgr.get_connection_targets.assert_awaited_once_with(
                connection_id=connection_id
            )
            assert message.target_list is conn_mgr.get_connection_targets.return_value

            mock_outbound_mgr.return_value.enqueue_message.assert_called_once_with(
                conductor.root_profile, message
            )

    async def test_outbound_message_handler_with_verkey_no_target(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            bus = conductor.root_profile.inject(EventBus)

            payload = "{}"
            message = OutboundMessage(
                payload=payload, reply_to_verkey=TestDIDs.test_verkey
            )

            status = await conductor.outbound_message_router(
                conductor.root_profile,
                message,
                inbound=mock.MagicMock(
                    receipt=mock.MagicMock(recipient_verkey=TestDIDs.test_verkey)
                ),
            )

            assert status == OutboundSendStatus.QUEUED_FOR_DELIVERY
            assert bus.events
            assert bus.events[0][1].topic == status.topic
            assert bus.events[0][1].payload == message

            mock_outbound_mgr.return_value.enqueue_message.assert_called_once_with(
                conductor.root_profile, message
            )

    async def test_handle_nots(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", mock.MagicMock()
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value = mock.MagicMock(
                setup=mock.CoroutineMock(),
                enqueue_message=mock.CoroutineMock(),
            )

            payload = "{}"
            message = OutboundMessage(
                payload=payload,
                connection_id="dummy-conn-id",
                reply_to_verkey=TestDIDs.test_verkey,
            )

            await conductor.setup()

            assert conductor.root_profile
            conductor.handle_not_returned(conductor.root_profile, message)

            mock_conn_mgr = mock.MagicMock()
            conductor.context.injector.bind_instance(BaseConnectionManager, mock_conn_mgr)
            with (
                mock.patch.object(
                    conductor.dispatcher, "run_task", mock.MagicMock()
                ) as mock_run_task,
            ):
                mock_run_task.side_effect = test_module.BaseConnectionManagerError()
                await conductor.queue_outbound(conductor.root_profile, message)

                message.connection_id = None
                await conductor.queue_outbound(conductor.root_profile, message)
                mock_run_task.assert_called_once()

    async def test_handle_outbound_queue(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        payload = "{}"
        message = OutboundMessage(
            payload=payload,
            connection_id="dummy-conn-id",
            target=mock.MagicMock(endpoint="endpoint"),
            reply_to_verkey=TestDIDs.test_verkey,
        )
        test_profile = await create_test_profile(None, await builder.build_context())

        with mock.patch.object(
            test_module,
            "wallet_config",
            return_value=(
                test_profile,
                DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
            ),
        ):
            await conductor.setup()

        await conductor.queue_outbound(conductor.root_profile, message)

    async def test_handle_not_returned_ledger_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
        with (
            mock.patch.object(
                conductor.dispatcher, "run_task", mock.MagicMock()
            ) as mock_dispatch_run,
            mock.patch.object(conductor, "queue_outbound", mock.MagicMock()),
        ):
            mock_dispatch_run.side_effect = test_module.LedgerConfigError(
                "No such ledger"
            )

            payload = "{}"
            message = OutboundMessage(
                payload=payload,
                connection_id="dummy-conn-id",
                reply_to_verkey=TestDIDs.test_verkey,
            )

            with self.assertRaises(test_module.LedgerConfigError):
                conductor.handle_not_returned(conductor.root_profile, message)

            mock_dispatch_run.assert_called_once()

    async def test_queue_outbound_ledger_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        conn_mgr = mock.MagicMock()
        conductor.context.injector.bind_instance(BaseConnectionManager, conn_mgr)
        with (
            mock.patch.object(
                conductor.dispatcher, "run_task", mock.MagicMock()
            ) as mock_dispatch_run,
            mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger,
        ):
            mock_dispatch_run.side_effect = test_module.LedgerConfigError(
                "No such ledger"
            )

            payload = "{}"
            message = OutboundMessage(
                payload=payload,
                connection_id="dummy-conn-id",
                reply_to_verkey=TestDIDs.test_verkey,
            )

            with self.assertRaises(test_module.LedgerConfigError):
                await conductor.queue_outbound(conductor.root_profile, message)

            mock_dispatch_run.assert_called_once()
            mock_logger.error.assert_called_once()

    async def test_admin(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"admin.enabled": "1"})
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
        admin = conductor.context.inject(BaseAdminServer)
        assert admin is conductor.admin_server

        with (
            mock.patch.object(admin, "start", autospec=True) as admin_start,
            mock.patch.object(admin, "stop", autospec=True) as admin_stop,
        ):
            await conductor.start()
            admin_start.assert_awaited_once_with()

            await conductor.stop()
            admin_stop.assert_awaited_once_with()

    async def test_admin_startx(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings(
            {
                "admin.enabled": "1",
                "debug.print_invitation": True,
                "debug.print_connections_invitation": True,
                "wallet.type": "askar",
            }
        )
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
        admin = conductor.context.inject(BaseAdminServer)
        assert admin is conductor.admin_server

        with (
            mock.patch.object(admin, "start", autospec=True) as admin_start,
            mock.patch.object(admin, "stop", autospec=True) as admin_stop,
            mock.patch.object(test_module, "OutOfBandManager") as oob_mgr,
            mock.patch.object(test_module, "BaseConnectionManager") as conn_mgr,
        ):
            admin_start.side_effect = KeyError("trouble")
            oob_mgr.return_value.create_invitation = mock.CoroutineMock(
                side_effect=KeyError("double trouble")
            )
            conn_mgr.return_value.create_invitation = mock.CoroutineMock(
                side_effect=KeyError("triple trouble")
            )
            await conductor.start()
            admin_start.assert_awaited_once_with()

            await conductor.stop()
            admin_stop.assert_awaited_once_with()

    async def test_setup_collector(self):
        builder: ContextBuilder = StubCollectorContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

    async def test_start_static(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"debug.test_suite_endpoint": True})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            mock_mgr.return_value.create_static_connection = mock.CoroutineMock(
                return_value=(None, None, mock.MagicMock())
            )
            await conductor.start()
            mock_mgr.return_value.create_static_connection.assert_awaited_once()

    async def test_start_x_in(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"debug.test_suite_endpoint": True})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(test_module, "InboundTransportManager") as mock_intx_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_intx_mgr.return_value = mock.MagicMock(
                setup=mock.CoroutineMock(),
                start=mock.CoroutineMock(side_effect=KeyError("trouble")),
            )
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            mock_mgr.return_value.create_static_connection = mock.CoroutineMock()
            with self.assertRaises(KeyError):
                await conductor.start()

    async def test_start_x_out_a(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"debug.test_suite_endpoint": True})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(test_module, "OutboundTransportManager") as mock_outx_mgr,
        ):
            mock_outx_mgr.return_value = mock.MagicMock(
                setup=mock.CoroutineMock(),
                start=mock.CoroutineMock(side_effect=KeyError("trouble")),
                registered_transports={"test": mock.MagicMock(schemes=["http"])},
            )
            await conductor.setup()
            mock_mgr.return_value.create_static_connection = mock.CoroutineMock()
            with self.assertRaises(KeyError):
                await conductor.start()

    async def test_start_x_out_b(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"debug.test_suite_endpoint": True})
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(test_module, "OutboundTransportManager") as mock_outx_mgr,
        ):
            mock_outx_mgr.return_value = mock.MagicMock(
                setup=mock.CoroutineMock(),
                start=mock.CoroutineMock(side_effect=KeyError("trouble")),
                stop=mock.CoroutineMock(),
                registered_transports={},
                enqueue_message=mock.CoroutineMock(),
            )
            await conductor.setup()
            mock_mgr.return_value.create_static_connection = mock.CoroutineMock()
            with self.assertRaises(KeyError):
                await conductor.start()

    async def test_dispatch_complete_non_fatal_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)

        message_body = "{}"
        receipt = MessageReceipt(direct_response_mode="snail mail")
        message = InboundMessage(message_body, receipt)
        exc = StorageNotFoundError("sample exception")
        mock_task = mock.MagicMock(
            exc_info=(type(exc), exc, exc.__traceback__),
            ident="abc",
            timing={
                "queued": 1234567890,
                "unqueued": 1234567899,
                "started": 1234567901,
                "ended": 1234567999,
            },
        )

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with mock.patch.object(
            conductor.admin_server, "notify_fatal_error", mock.MagicMock()
        ) as mock_notify:
            conductor.dispatch_complete(message, mock_task)
            mock_notify.assert_not_called()

    async def test_dispatch_complete_ledger_error_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings_admin)
        conductor = test_module.Conductor(builder)

        message_body = "{}"
        receipt = MessageReceipt(direct_response_mode="snail mail")
        message = InboundMessage(message_body, receipt)
        exc = test_module.LedgerTransactionError("Ledger is wobbly")
        mock_task = mock.MagicMock(
            exc_info=(type(exc), exc, exc.__traceback__),
            ident="abc",
            timing={
                "queued": 1234567890,
                "unqueued": 1234567899,
                "started": 1234567901,
                "ended": 1234567999,
            },
        )

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger:
            conductor.dispatch_complete(message, mock_task)
            mock_logger.error.assert_called_once()

    async def test_clear_default_mediator(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"mediation.clear": True})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with mock.patch.object(
            test_module,
            "MediationManager",
            return_value=mock.MagicMock(clear_default_mediator=mock.CoroutineMock()),
        ) as mock_mgr:
            await conductor.start()
            await conductor.stop()
            mock_mgr.return_value.clear_default_mediator.assert_called_once()

    async def test_set_default_mediator(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"mediation.default_id": "test-id"})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with (
            mock.patch.object(
                test_module,
                "MediationManager",
                return_value=mock.MagicMock(
                    set_default_mediator_by_id=mock.CoroutineMock()
                ),
            ) as mock_mgr,
            mock.patch.object(MediationRecord, "retrieve_by_id", mock.CoroutineMock()),
            mock.patch.object(
                test_module,
                "LOGGER",
                mock.MagicMock(
                    exception=mock.MagicMock(
                        side_effect=Exception("This method should not have been called")
                    )
                ),
            ),
        ):
            await conductor.start()
            await conductor.stop()
            mock_mgr.return_value.set_default_mediator_by_id.assert_called_once()

    async def test_set_default_mediator_x(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"mediation.default_id": "test-id"})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        with (
            mock.patch.object(
                MediationRecord,
                "retrieve_by_id",
                mock.CoroutineMock(side_effect=Exception()),
            ),
            mock.patch.object(test_module, "LOGGER") as mock_logger,
        ):
            await conductor.start()
            await conductor.stop()
            mock_logger.exception.assert_called_once()

    async def test_webhook_router(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings(
            {"debug.print_invitation": True, "invite_base_url": "http://localhost"}
        )
        conductor = test_module.Conductor(builder)

        test_topic = "test-topic"
        test_payload = {"test": "payload"}
        test_endpoint = "http://example"
        test_attempts = 2

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
        with mock.patch.object(
            conductor.outbound_transport_manager, "enqueue_webhook"
        ) as mock_enqueue:
            conductor.webhook_router(
                test_topic, test_payload, test_endpoint, test_attempts
            )
            mock_enqueue.assert_called_once_with(
                test_topic, test_payload, test_endpoint, test_attempts, None
            )

        # swallow error
        with mock.patch.object(
            conductor.outbound_transport_manager,
            "enqueue_webhook",
            side_effect=OutboundDeliveryError,
        ) as mock_enqueue:
            conductor.webhook_router(
                test_topic, test_payload, test_endpoint, test_attempts
            )
            mock_enqueue.assert_called_once_with(
                test_topic, test_payload, test_endpoint, test_attempts, None
            )

    async def test_shutdown_multitenant_profiles(self):
        builder: ContextBuilder = StubContextBuilder(
            {**self.test_settings, "multitenant.enabled": True}
        )
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            multitenant_mgr = conductor.context.inject(BaseMultitenantManager)
            assert isinstance(multitenant_mgr, MultitenantManager)

            multitenant_mgr._profiles.put(
                "test1",
                mock.MagicMock(close=mock.CoroutineMock()),
            )
            multitenant_mgr._profiles.put(
                "test2",
                mock.MagicMock(close=mock.CoroutineMock()),
            )

            await conductor.stop()

            multitenant_mgr._profiles.profiles["test1"].close.assert_called_once_with()
            multitenant_mgr._profiles.profiles["test2"].close.assert_called_once_with()


def get_invite_store_mock(
    invite_string: str, invite_already_used: bool = False
) -> mock.MagicMock:
    unused_invite = MediationInviteRecord(invite_string, invite_already_used)
    used_invite = MediationInviteRecord(invite_string, used=True)

    return mock.MagicMock(
        get_mediation_invite_record=mock.CoroutineMock(return_value=unused_invite),
        mark_default_invite_as_used=mock.CoroutineMock(return_value=used_invite),
    )


class TestConductorMediationSetup(IsolatedAsyncioTestCase, Config):
    """
    Test related with setting up mediation from given arguments or stored invitation.
    """

    def __get_mediator_config(
        self, invite_string: str, connections_invite: bool = False
    ) -> ContextBuilder:
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"mediation.invite": invite_string})
        if connections_invite:
            builder.update_settings({"mediation.connections_invite": True})

        return builder

    @mock.patch.object(
        test_module,
        "MediationInviteStore",
        return_value=get_invite_store_mock("test-invite"),
    )
    @mock.patch.object(test_module.InvitationMessage, "from_url")
    async def test_mediator_invitation_0434(self, mock_from_url, _):
        builder = self.__get_mediator_config("test-invite", True)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        conductor.root_profile.context.update_settings(
            {"mediation.connections_invite": False}
        )
        conn_record = ConnRecord(
            invitation_key="3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
            their_label="Hello",
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            alias="Bob",
        )
        conn_record.accept = ConnRecord.ACCEPT_MANUAL
        await conn_record.save(await conductor.root_profile.session())
        invitation = test_module.InvitationMessage()
        oob_record = OobRecord(
            invitation=invitation,
            invi_msg_id=invitation._id,
            role=OobRecord.ROLE_RECEIVER,
            connection_id=conn_record.connection_id,
            state=OobRecord.STATE_INITIAL,
        )

        with (
            mock.patch.object(
                test_module,
                "OutOfBandManager",
                mock.MagicMock(
                    return_value=mock.MagicMock(
                        receive_invitation=mock.CoroutineMock(return_value=oob_record)
                    )
                ),
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "upgrade_wallet_to_anoncreds_if_requested",
                return_value=False,
            ),
        ):
            assert not conductor.root_profile.settings["mediation.connections_invite"]
            await conductor.start()
            await conductor.stop()
            mock_from_url.assert_called_once_with("test-invite")
            mock_mgr.return_value.receive_invitation.assert_called_once()

    @mock.patch.object(test_module, "MediationInviteStore")
    @mock.patch.object(test_module, "BaseConnectionManager")
    async def test_mediation_invitation_should_not_create_connection_for_old_invitation(
        self, patched_connection_manager, patched_invite_store
    ):
        # given
        invite_string = "test-invite"

        builder = self.__get_mediator_config("test-invite", True)
        conductor = test_module.Conductor(builder)
        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

        invite_store_mock = get_invite_store_mock(invite_string, True)
        patched_invite_store.return_value = invite_store_mock

        connection_manager_mock = mock.MagicMock(receive_invitation=mock.CoroutineMock())
        patched_connection_manager.return_value = connection_manager_mock
        # when
        await conductor.start()
        await conductor.stop()

        # then
        invite_store_mock.get_mediation_invite_record.assert_called_with(invite_string)
        connection_manager_mock.receive_invitation.assert_not_called()

    async def test_setup_ledger_both_multiple_and_base(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"ledger.genesis_transactions": "..."})
        builder.update_settings(
            {
                "ledger.ledger_config_list": [
                    {
                        "id": "sovrinMain",
                        "is_production": True,
                        "is_write": True,
                    },
                ]
            }
        )
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())
        test_profile.context.injector.bind_instance(
            BaseLedger, mock.MagicMock(BaseLedger, autospec=True)
        )

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module,
                "load_multiple_genesis_transactions_from_config",
                mock.CoroutineMock(),
            ) as mock_multiple_genesis_load,
            mock.patch.object(
                test_module, "get_genesis_transactions", mock.CoroutineMock()
            ) as mock_genesis_load,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(test_module, "ledger_config"),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            mock_multiple_genesis_load.assert_called_once()
            mock_genesis_load.assert_called_once()

    async def test_setup_ledger_only_base(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        builder.update_settings({"ledger.genesis_transactions": "..."})
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "get_genesis_transactions", mock.CoroutineMock()
            ) as mock_genesis_load,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(test_module, "ledger_config"),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()
            mock_genesis_load.assert_called_once()

    async def test_startup_storage_type_anoncreds_and_config_askar_re_calls_setup(self):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module,
                "upgrade",
                mock.CoroutineMock(),
            ),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }

            async with test_profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(
                    StorageRecord(RECORD_TYPE_ACAPY_STORAGE_TYPE, "askar-anoncreds")
                )

            await conductor.setup()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}
            with mock.patch.object(test_module.Conductor, "setup") as mock_setup:
                mock_setup.return_value = None
                with self.assertRaises(Exception):
                    await conductor.start()
                assert mock_setup.called

            # await conductor.stop()

    async def test_startup_storage_type_does_not_exist_and_existing_agent_then_set_to_askar(
        self,
    ):
        builder: ContextBuilder = StubContextBuilder(self.test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(None, await builder.build_context())

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(
                    return_value=["v0.7.4", "0.7.5", "v0.8.0-rc1", "v8.0.0", "v0.8.1-rc2"]
                ),
            ),
            mock.patch.object(
                test_module,
                "upgrade",
                mock.CoroutineMock(),
            ),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}

            await conductor.start()

            async with test_profile.session() as session:
                storage = session.inject(BaseStorage)
                record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
                )
                assert record.value == "askar"

            await conductor.stop()

    async def test_startup_storage_type_does_not_exist_and_new_anoncreds_agent(
        self,
    ):
        test_settings = {
            "admin.webhook_urls": ["http://sample.webhook.ca"],
            "wallet.type": "askar-anoncreds",
        }
        builder: ContextBuilder = StubContextBuilder(test_settings)
        conductor = test_module.Conductor(builder)

        test_profile = await create_test_profile(
            test_settings, await builder.build_context()
        )

        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                return_value=(
                    test_profile,
                    DIDInfo("did", "verkey", metadata={}, method=SOV, key_type=ED25519),
                ),
            ),
            mock.patch.object(
                test_module, "InboundTransportManager", autospec=True
            ) as mock_inbound_mgr,
            mock.patch.object(
                test_module, "OutboundTransportManager", autospec=True
            ) as mock_outbound_mgr,
            mock.patch.object(
                test_module,
                "get_upgrade_version_list",
                mock.MagicMock(
                    return_value=["v0.7.4", "0.7.5", "v0.8.0-rc1", "v8.0.0", "v0.8.1-rc2"]
                ),
            ),
            mock.patch.object(
                test_module,
                "upgrade",
                mock.CoroutineMock(),
            ),
        ):
            mock_outbound_mgr.return_value.registered_transports = {
                "test": mock.MagicMock(schemes=["http"])
            }
            await conductor.setup()

            mock_inbound_mgr.return_value.registered_transports = {}
            mock_outbound_mgr.return_value.registered_transports = {}

            await conductor.start()
            async with test_profile.session() as session:
                storage = session.inject(BaseStorage)
                record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
                )
                assert record.value == "askar-anoncreds"

            await conductor.stop()



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import TRANSACTION_REQUEST
from ..transaction_request import TransactionRequest


class TestConfig:
    test_transaction_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
    test_signature_request = {
        "author_goal_code": "transaction.ledger.write",
        "context": "did:sov",
        "method": "add-signature",
        "signature_type": "<requested signature type>",
        "signer_goal_code": "transaction.endorse",
    }
    test_timing = {"expires_time": "1597708800"}
    test_transaction_type = "http://didcomm.org/sign-attachment/%VER/signature-request"
    test_messages_attach = {
        "@id": "143c458d-1b1c-40c7-ab85-4d16808ddf0a",
        "mime-type": "application/json",
        "data": {
            "json": {
                "endorser": "V4SGRU86Z58d6TV7PBUe6f",
                "identifier": "LjgpST2rjsoxYegQDRm7EL",
                "operation": {
                    "data": {
                        "attr_names": ["first_name", "last_name"],
                        "name": "test_schema",
                        "version": "2.1",
                    },
                    "type": "101",
                },
                "protocolVersion": 2,
                "reqId": 1597766666168851000,
                "signatures": {
                    "LjgpST2rjsox": "4uq1mUATKMn6Y9sTaGWyuPgjUEw5UBysWNbfSqCfnbm1Vnfw"
                },
                "taaAcceptance": {
                    "mechanism": "manual",
                    "taaDigest": "f50feca75664270842bd4202c2d6f23e4c6a7e0fc2feb9f62",
                    "time": 1597708800,
                },
            }
        },
    }


class TestTransactionRequest(TestCase, TestConfig):
    def setUp(self):
        self.transaction_request = TransactionRequest(
            transaction_id=self.test_transaction_id,
            signature_request=self.test_signature_request,
            timing=self.test_timing,
            transaction_type=self.test_transaction_type,
            messages_attach=self.test_messages_attach,
        )

    def test_init(self):
        """Test initialization."""
        assert self.transaction_request.transaction_id == self.test_transaction_id
        assert self.transaction_request.signature_request == self.test_signature_request
        assert self.transaction_request.timing == self.test_timing
        assert self.transaction_request.transaction_type == self.test_transaction_type
        assert self.transaction_request.messages_attach == self.test_messages_attach

    def test_type(self):
        """Test type."""
        assert self.transaction_request._type == DIDCommPrefix.qualify_current(
            TRANSACTION_REQUEST
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_request.TransactionRequestSchema.load"
    )
    def test_deserialize(self, mock_transaction_request_schema_load):
        """
        Test deserialization.
        """
        obj = self.transaction_request

        transaction_request = TransactionRequest.deserialize(obj)
        mock_transaction_request_schema_load.assert_called_once_with(obj)

        assert transaction_request is mock_transaction_request_schema_load.return_value

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_request.TransactionRequestSchema.dump"
    )
    def test_serialize(self, mock_transaction_request_schema_dump):
        """
        Test serialization.
        """
        transaction_request_dict = self.transaction_request.serialize()
        mock_transaction_request_schema_dump.assert_called_once_with(
            self.transaction_request
        )

        assert (
            transaction_request_dict is mock_transaction_request_schema_dump.return_value
        )


class TestTransactionRequestSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test transaction request schema."""

    async def test_make_model(self):
        transaction_request = TransactionRequest(
            transaction_id=self.test_transaction_id,
            signature_request=self.test_signature_request,
            timing=self.test_timing,
            transaction_type=self.test_transaction_type,
            messages_attach=self.test_messages_attach,
        )
        data = transaction_request.serialize()
        model_instance = TransactionRequest.deserialize(data)
        assert type(model_instance) is type(transaction_request)



"""Represents an action menu."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import MENU, PROTOCOL_PACKAGE
from ..models.menu_option import MenuOption, MenuOptionSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.menu_handler.MenuHandler"


class Menu(AgentMessage):
    """Class representing an action menu."""

    class Meta:
        """Metadata for an action menu."""

        handler_class = HANDLER_CLASS
        message_type = MENU
        schema_class = "MenuSchema"

    def __init__(
        self,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        errormsg: Optional[str] = None,
        options: Sequence[MenuOption] = None,
        **kwargs,
    ):
        """Initialize a menu object.

        Args:
            title: The menu title
            description: Introductory text for the menu
            errormsg: An optional error message to display
            options: A sequence of menu options
            kwargs: Additional keyword arguments
        """
        super().__init__(**kwargs)
        self.title = title
        self.description = description
        self.options = list(options) if options else []


class MenuSchema(AgentMessageSchema):
    """Menu schema class."""

    class Meta:
        """Menu schema metadata."""

        model_class = Menu
        unknown = EXCLUDE

    title = fields.Str(
        required=False, metadata={"description": "Menu title", "example": "My Menu"}
    )
    description = fields.Str(
        required=False,
        metadata={
            "description": "Introductory text for the menu",
            "example": "This menu presents options",
        },
    )
    errormsg = fields.Str(
        required=False,
        metadata={
            "description": "An optional error message to display in menu header",
            "example": "Error: item not found",
        },
    )
    options = fields.List(
        fields.Nested(MenuOptionSchema),
        required=True,
        metadata={"description": "List of menu options"},
    )



"""V2.0 issue-credential indy credential format handler."""

import asyncio
import json
import logging
from typing import Mapping, Optional, Tuple

from marshmallow import RAISE

from ......askar.profile_anon import AskarAnonCredsProfile
from ......cache.base import BaseCache
from ......core.profile import Profile
from ......indy.holder import IndyHolder, IndyHolderError
from ......indy.issuer import IndyIssuer, IndyIssuerRevocationRegistryFullError
from ......indy.models.cred import IndyCredentialSchema
from ......indy.models.cred_abstract import IndyCredAbstractSchema
from ......indy.models.cred_request import IndyCredRequestSchema
from ......ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    GET_SCHEMA,
    IndyLedgerRequestsExecutor,
)
from ......messaging.credential_definitions.util import (
    CRED_DEF_SENT_RECORD_TYPE,
    CredDefQueryStringSchema,
)
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......multitenant.base import BaseMultitenantManager
from ......revocation.indy import IndyRevocation
from ......revocation.models.issuer_cred_rev_record import IssuerCredRevRecord
from ......revocation.models.revocation_registry import RevocationRegistry
from ......storage.base import BaseStorage
from ...message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ...messages.cred_format import V20CredFormat
from ...messages.cred_issue import V20CredIssue
from ...messages.cred_offer import V20CredOffer
from ...messages.cred_proposal import V20CredProposal
from ...messages.cred_request import V20CredRequest
from ...models.cred_ex_record import V20CredExRecord
from ...models.detail.indy import V20CredExRecordIndy
from ..anoncreds.handler import AnonCredsCredFormatHandler
from ..handler import CredFormatAttachment, V20CredFormatError, V20CredFormatHandler

LOGGER = logging.getLogger(__name__)


class IndyCredFormatHandler(V20CredFormatHandler):
    """Indy credential format handler."""

    format = V20CredFormat.Format.INDY
    anoncreds_handler = None

    def __init__(self, profile: Profile):
        """Shim initialization to check for new AnonCreds library."""
        super().__init__(profile)

        # Temporary shim while the new anoncreds library integration is in progress
        wallet_type = profile.settings.get_value("wallet.type")
        if wallet_type == "askar-anoncreds":
            self.anoncreds_handler = AnonCredsCredFormatHandler(profile)

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping):
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.


        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to validate

        Raises:
            Exception: When the data is not valid.

        """
        mapping = {
            CRED_20_PROPOSAL: CredDefQueryStringSchema,
            CRED_20_OFFER: IndyCredAbstractSchema,
            CRED_20_REQUEST: IndyCredRequestSchema,
            CRED_20_ISSUE: IndyCredentialSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=RAISE).load(attachment_data)

    async def get_detail_record(self, cred_ex_id: str) -> V20CredExRecordIndy:
        """Retrieve credential exchange detail record by cred_ex_id."""

        async with self.profile.session() as session:
            records = await IndyCredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )

        if len(records) > 1:
            LOGGER.warning(
                "Cred ex id %s has %d %s detail records: should be 1",
                cred_ex_id,
                len(records),
                IndyCredFormatHandler.format.api,
            )
        return records[0] if records else None

    async def _check_uniqueness(self, cred_ex_id: str):
        """Raise exception on evidence that cred ex already has cred issued to it."""
        async with self.profile.session() as session:
            exist = await IndyCredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )
        if exist:
            raise V20CredFormatError(
                f"{IndyCredFormatHandler.format.api} detail record already "
                f"exists for cred ex id {cred_ex_id}"
            )

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """

        return ATTACHMENT_FORMAT[message_type][IndyCredFormatHandler.format.api]

    def get_format_data(self, message_type: str, data: dict) -> CredFormatAttachment:
        """Get credential format and attachment objects for use in cred ex messages.

        Returns a tuple of both credential format and attachment decorator for use
        in credential exchange messages. It looks up the correct format identifier and
        encodes the data as a base64 attachment.

        Args:
            message_type (str): The message type for which to return the cred format.
                Should be one of the message types defined in the message types file
            data (dict): The data to include in the attach decorator

        Returns:
            CredFormatAttachment: Credential format and attachment data objects

        """

        return (
            V20CredFormat(
                attach_id=IndyCredFormatHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(data, ident=IndyCredFormatHandler.format.api),
        )

    async def _match_sent_cred_def_id(self, tag_query: Mapping[str, str]) -> str:
        """Return most recent matching id of cred def that agent sent to ledger."""

        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            found = await storage.find_all_records(
                type_filter=CRED_DEF_SENT_RECORD_TYPE, tag_query=tag_query
            )
        if not found:
            raise V20CredFormatError(
                f"Issuer has no operable cred def for proposal spec {tag_query}"
            )
        return max(found, key=lambda r: int(r.tags["epoch"])).tags["cred_def_id"]

    async def create_proposal(
        self, cred_ex_record: V20CredExRecord, proposal_data: Mapping[str, str]
    ) -> Tuple[V20CredFormat, AttachDecorator]:
        """Create indy credential proposal."""
        # Create the proposal with the anoncreds handler if agent is anoncreds capable
        if self.anoncreds_handler:
            return await self.anoncreds_handler.create_proposal(
                cred_ex_record,
                proposal_data,
            )

        if proposal_data is None:
            proposal_data = {}

        return self.get_format_data(CRED_20_PROPOSAL, proposal_data)

    async def receive_proposal(
        self, cred_ex_record: V20CredExRecord, cred_proposal_message: V20CredProposal
    ) -> None:
        """Receive indy credential proposal.

        No custom handling is required for this step.
        """

    async def create_offer(
        self, cred_proposal_message: V20CredProposal
    ) -> CredFormatAttachment:
        """Create indy credential offer."""

        if isinstance(self.profile, AskarAnonCredsProfile):
            raise V20CredFormatError(
                "This issuer is anoncreds capable. Please use the anoncreds format."
            )

        issuer = self.profile.inject(IndyIssuer)
        cache = self.profile.inject_or(BaseCache)

        cred_def_id = await self._match_sent_cred_def_id(
            cred_proposal_message.attachment(IndyCredFormatHandler.format)
        )

        async def _create():
            offer_json = await issuer.create_credential_offer(cred_def_id)
            return json.loads(offer_json)

        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)
            schema = await ledger.get_schema(schema_id)
        schema_attrs = set(schema["attrNames"])
        preview_attrs = set(cred_proposal_message.credential_preview.attr_dict())
        if preview_attrs != schema_attrs:
            raise V20CredFormatError(
                f"Preview attributes {preview_attrs} "
                f"mismatch corresponding schema attributes {schema_attrs}"
            )

        cred_offer = None
        cache_key = f"credential_offer::{cred_def_id}"

        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_offer = entry.result
                else:
                    cred_offer = await _create()
                    await entry.set_result(cred_offer, 3600)
        if not cred_offer:
            cred_offer = await _create()

        return self.get_format_data(CRED_20_OFFER, cred_offer)

    async def receive_offer(
        self, cred_ex_record: V20CredExRecord, cred_offer_message: V20CredOffer
    ) -> None:
        """Receive indy credential offer."""

    async def create_cred_request_result(self, cred_offer, holder_did, cred_def_id):
        """Create credential request result."""
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            cred_def = await ledger.get_credential_definition(cred_def_id)

        holder = self.profile.inject(IndyHolder)
        request_json, metadata_json = await holder.create_credential_request(
            cred_offer, cred_def, holder_did
        )

        return {
            "request": json.loads(request_json),
            "metadata": json.loads(metadata_json),
        }

    async def create_request(
        self, cred_ex_record: V20CredExRecord, request_data: Optional[Mapping] = None
    ) -> CredFormatAttachment:
        """Create indy credential request."""

        # Create the request with the anoncreds handler if agent is anoncreds capable
        if self.anoncreds_handler:
            return await self.anoncreds_handler.create_request(
                cred_ex_record,
                request_data,
            )

        if cred_ex_record.state != V20CredExRecord.STATE_OFFER_RECEIVED:
            raise V20CredFormatError(
                "Indy issue credential format cannot start from credential request"
            )

        await self._check_uniqueness(cred_ex_record.cred_ex_id)

        holder_did = request_data.get("holder_did") if request_data else None
        cred_offer = cred_ex_record.cred_offer.attachment(IndyCredFormatHandler.format)

        if "nonce" not in cred_offer:
            raise V20CredFormatError("Missing nonce in credential offer")

        nonce = cred_offer["nonce"]
        cred_def_id = cred_offer["cred_def_id"]

        cache_key = f"credential_request::{cred_def_id}::{holder_did}::{nonce}"
        cred_req_result = None
        cache = self.profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_req_result = entry.result
                else:
                    cred_req_result = await self.create_cred_request_result(
                        cred_offer, holder_did, cred_def_id
                    )
                    await entry.set_result(cred_req_result, 3600)
        if not cred_req_result:
            cred_req_result = await self.create_cred_request_result(
                cred_offer, holder_did, cred_def_id
            )

        detail_record = V20CredExRecordIndy(
            cred_ex_id=cred_ex_record.cred_ex_id,
            cred_request_metadata=cred_req_result["metadata"],
        )

        async with self.profile.session() as session:
            await detail_record.save(session, reason="create v2.0 credential request")

        return self.get_format_data(CRED_20_REQUEST, cred_req_result["request"])

    async def receive_request(
        self, cred_ex_record: V20CredExRecord, cred_request_message: V20CredRequest
    ) -> None:
        """Receive indy credential request."""
        # Receive the request with the anoncreds handler if agent is anoncreds capable
        if self.anoncreds_handler:
            return await self.anoncreds_handler.receive_request(
                cred_ex_record,
                cred_request_message,
            )

        if not cred_ex_record.cred_offer:
            raise V20CredFormatError(
                "Indy issue credential format cannot start from credential request"
            )

    async def _issue_credential_retry(
        self,
        retries: int,
        cred_def_id: str,
        schema: dict,
        cred_offer: dict,
        cred_request: dict,
        cred_values: dict[str, str],
        revocable: bool,
        issuer: IndyIssuer,
    ) -> tuple[Optional[dict], Optional[str], Optional[str]]:
        for _ in range(max(retries, 1)):
            if revocable:
                revoc = IndyRevocation(self.profile)
                registry_info = await revoc.get_or_create_active_registry(cred_def_id)
                if not registry_info:
                    continue
                del revoc
                issuer_rev_reg, rev_reg = registry_info
                rev_reg_id = issuer_rev_reg.revoc_reg_id
                tails_path = rev_reg.tails_local_path
            else:
                rev_reg_id = None
                tails_path = None

            try:
                (cred_json, cred_rev_id) = await issuer.create_credential(
                    schema,
                    cred_offer,
                    cred_request,
                    cred_values,
                    rev_reg_id,
                    tails_path,
                )
            except IndyIssuerRevocationRegistryFullError:
                # unlucky, another instance filled the registry first
                continue

            if revocable and rev_reg.max_creds <= int(cred_rev_id):
                revoc = IndyRevocation(self.profile)
                await revoc.handle_full_registry(rev_reg_id)
                del revoc

            result = self.get_format_data(CRED_20_ISSUE, json.loads(cred_json))
            if result:
                return result, rev_reg_id, cred_rev_id

            LOGGER.info(
                "Waiting 2s before retrying credential issuance for cred def '%s'",
                cred_def_id,
            )
            await asyncio.sleep(2)

        return None, None, None

    async def _get_ledger_for_schema(self, schema_id: str):
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        return (
            await ledger_exec_inst.get_ledger_for_identifier(
                schema_id,
                txn_record_type=GET_SCHEMA,
            )
        )[1]

    async def issue_credential(
        self, cred_ex_record: V20CredExRecord, retries: int = 5
    ) -> CredFormatAttachment:
        """Issue indy credential."""
        # Temporary shim while the new anoncreds library integration is in progress
        if self.anoncreds_handler:
            return await self.anoncreds_handler.issue_credential(cred_ex_record, retries)

        await self._check_uniqueness(cred_ex_record.cred_ex_id)

        cred_offer = cred_ex_record.cred_offer.attachment(IndyCredFormatHandler.format)
        from ..anoncreds.handler import AnonCredsCredFormatHandler

        cred_request = cred_ex_record.cred_request.attachment(
            IndyCredFormatHandler.format
        ) or cred_ex_record.cred_request.attachment(AnonCredsCredFormatHandler.format)
        cred_values = cred_ex_record.cred_offer.credential_preview.attr_dict(decode=False)
        schema_id = cred_offer["schema_id"]
        cred_def_id = cred_offer["cred_def_id"]

        issuer = self.profile.inject(IndyIssuer)
        ledger = await self._get_ledger_for_schema(schema_id)

        async with ledger:
            schema = await ledger.get_schema(schema_id)
            cred_def = await ledger.get_credential_definition(cred_def_id)

        revocable = True if cred_def["value"].get("revocation") else False

        result, rev_reg_id, cred_rev_id = await self._issue_credential_retry(
            retries,
            cred_def_id,
            schema,
            cred_offer,
            cred_request,
            cred_values,
            revocable,
            issuer,
        )

        if not result:
            raise V20CredFormatError(
                f"Cred def '{cred_def_id}' has no active revocation registry"
            )

        async with self._profile.transaction() as txn:
            detail_record = V20CredExRecordIndy(
                cred_ex_id=cred_ex_record.cred_ex_id,
                rev_reg_id=rev_reg_id,
                cred_rev_id=cred_rev_id,
            )
            await detail_record.save(txn, reason="v2.0 issue credential")

            if revocable and cred_rev_id:
                issuer_cr_rec = IssuerCredRevRecord(
                    state=IssuerCredRevRecord.STATE_ISSUED,
                    cred_ex_id=cred_ex_record.cred_ex_id,
                    cred_ex_version=IssuerCredRevRecord.VERSION_2,
                    rev_reg_id=rev_reg_id,
                    cred_rev_id=cred_rev_id,
                )
                await issuer_cr_rec.save(
                    txn,
                    reason=(
                        "Created issuer cred rev record for "
                        f"rev reg id {rev_reg_id}, index {cred_rev_id}"
                    ),
                )
            await txn.commit()

        return result

    async def receive_credential(
        self, cred_ex_record: V20CredExRecord, cred_issue_message: V20CredIssue
    ) -> None:
        """Receive indy credential.

        Validation is done in the store credential step.
        """

    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> None:
        """Store indy credential."""
        # Temporary shim while the new anoncreds library integration is in progress
        if hasattr(self, "anoncreds_handler") and self.anoncreds_handler:
            return await self.anoncreds_handler.store_credential(cred_ex_record, cred_id)

        from ..anoncreds.handler import AnonCredsCredFormatHandler

        cred = cred_ex_record.cred_issue.attachment(
            IndyCredFormatHandler.format
        ) or cred_ex_record.cred_issue.attachment(AnonCredsCredFormatHandler.format)

        rev_reg_def = None
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred["cred_def_id"],
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            cred_def = await ledger.get_credential_definition(cred["cred_def_id"])
            if cred.get("rev_reg_id"):
                rev_reg_def = await ledger.get_revoc_reg_def(cred["rev_reg_id"])

        holder = self.profile.inject(IndyHolder)
        cred_offer_message = cred_ex_record.cred_offer
        mime_types = None
        if cred_offer_message and cred_offer_message.credential_preview:
            mime_types = cred_offer_message.credential_preview.mime_types() or None

        if rev_reg_def:
            rev_reg = RevocationRegistry.from_definition(rev_reg_def, True)
            await rev_reg.get_or_fetch_local_tails_path()
        try:
            detail_record = await self.get_detail_record(cred_ex_record.cred_ex_id)
            if detail_record is None:
                raise V20CredFormatError(
                    f"No credential exchange {IndyCredFormatHandler.format.aries} "
                    f"detail record found for cred ex id {cred_ex_record.cred_ex_id}"
                )
            cred_id_stored = await holder.store_credential(
                cred_def,
                cred,
                detail_record.cred_request_metadata,
                mime_types,
                credential_id=cred_id,
                rev_reg_def=rev_reg_def,
            )

            detail_record.cred_id_stored = cred_id_stored
            detail_record.rev_reg_id = cred.get("rev_reg_id", None)
            detail_record.cred_rev_id = cred.get("cred_rev_id", None)

            async with self.profile.session() as session:
                # Store detail record, emit event
                await detail_record.save(
                    session, reason="store credential v2.0", event=True
                )
        except IndyHolderError as e:
            LOGGER.error(f"Error storing credential: {e.error_code} - {e.message}")
            raise e



"""Route manager.

Set up routing for newly formed connections.
"""

import logging
from abc import ABC, abstractmethod
from typing import List, NamedTuple, Optional

from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....messaging.responder import BaseResponder
from ....storage.error import StorageNotFoundError
from ....wallet.base import BaseWallet
from ....wallet.did_info import DIDInfo
from ....wallet.did_method import SOV
from ....wallet.key_type import ED25519
from ...routing.v1_0.models.route_record import RouteRecord
from .manager import MediationManager
from .messages.keylist_update import KeylistUpdate
from .models.mediation_record import MediationRecord
from .normalization import normalize_from_did_key, normalize_to_did_key

LOGGER = logging.getLogger(__name__)


class RouteManagerError(Exception):
    """Raised on error from route manager."""


class RoutingInfo(NamedTuple):
    """Routing info tuple containing routing keys and endpoint."""

    routing_keys: Optional[List[str]]
    endpoint: Optional[str]

    @classmethod
    def empty(cls):
        """Empty routing info."""
        return cls(routing_keys=None, endpoint=None)


class RouteManager(ABC):
    """Base Route Manager."""

    async def get_or_create_my_did(
        self, profile: Profile, conn_record: ConnRecord
    ) -> DIDInfo:
        """Create or retrieve DID info for a connection."""
        if not conn_record.my_did:
            async with profile.session() as session:
                wallet = session.inject(BaseWallet)

                LOGGER.debug(
                    "Creating new DID for connection %s", conn_record.connection_id
                )
                my_info = await wallet.create_local_did(SOV, ED25519)
                conn_record.my_did = my_info.did
                await conn_record.save(session, reason="Connection my did created")
        else:
            async with profile.session() as session:
                wallet = session.inject(BaseWallet)
                LOGGER.debug(
                    "Getting DID info for connection %s", conn_record.connection_id
                )
                my_info = await wallet.get_local_did(conn_record.my_did)

        return my_info

    def _validate_mediation_state(self, mediation_record: MediationRecord):
        """Perform mediation state validation."""
        if mediation_record.state != MediationRecord.STATE_GRANTED:
            LOGGER.error(
                "Mediation is not granted for mediation identified by %s",
                mediation_record.mediation_id,
            )
            raise RouteManagerError(
                "Mediation is not granted for mediation identified by "
                f"{mediation_record.mediation_id}"
            )

    async def mediation_records_for_connection(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_id: Optional[str] = None,
        or_default: bool = False,
    ) -> List[MediationRecord]:
        """Return relevant mediator for connection."""
        # TODO Support multiple mediators?
        if conn_record.connection_id:
            async with profile.session() as session:
                mediation_metadata = await conn_record.metadata_get(
                    session, MediationManager.METADATA_KEY, {}
                )
                mediation_id = (
                    mediation_metadata.get(MediationManager.METADATA_ID) or mediation_id
                )

        mediation_record = await self.mediation_record_if_id(
            profile, mediation_id, or_default
        )
        if mediation_record:
            await self.save_mediator_for_connection(
                profile, conn_record, mediation_record
            )
        return [mediation_record] if mediation_record else []

    async def mediation_record_if_id(
        self,
        profile: Profile,
        mediation_id: Optional[str] = None,
        or_default: bool = False,
    ):
        """Validate mediation and return record.

        If mediation_id is not None,
        validate mediation record state and return record
        else, return None
        """
        mediation_record = None
        if mediation_id:
            async with profile.session() as session:
                mediation_record = await MediationRecord.retrieve_by_id(
                    session, mediation_id
                )
        elif or_default:
            mediation_record = await MediationManager(profile).get_default_mediator()

        if mediation_record:
            self._validate_mediation_state(mediation_record)
        return mediation_record

    @abstractmethod
    async def _route_for_key(
        self,
        profile: Profile,
        recipient_key: str,
        mediation_record: Optional[MediationRecord] = None,
        *,
        skip_if_exists: bool = False,
        replace_key: Optional[str] = None,
    ) -> Optional[KeylistUpdate]:
        """Route a key."""

    async def route_connection_as_invitee(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_records: List[MediationRecord],
    ) -> Optional[KeylistUpdate]:
        """Set up routing for a new connection when we are the invitee."""
        LOGGER.debug("Routing connection as invitee")
        my_info = await self.get_or_create_my_did(profile, conn_record)
        # Only most destward mediator receives keylist updates
        mediation_record = mediation_records[0] if mediation_records else None
        return await self._route_for_key(
            profile, my_info.verkey, mediation_record, skip_if_exists=True
        )

    async def route_connection_as_inviter(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_records: List[MediationRecord],
    ) -> Optional[KeylistUpdate]:
        """Set up routing for a new connection when we are the inviter."""
        LOGGER.debug("Routing connection as inviter")
        my_info = await self.get_or_create_my_did(profile, conn_record)

        replace_key = conn_record.invitation_key
        async with profile.session() as session:
            wallet = session.inject(BaseWallet)
            public_did = await wallet.get_public_did()

        # Do not replace key, if it is public
        if public_did and public_did.verkey == conn_record.invitation_key:
            replace_key = None

        # Only most destward mediator receives keylist updates
        mediation_record = mediation_records[0] if mediation_records else None

        return await self._route_for_key(
            profile,
            my_info.verkey,
            mediation_record,
            replace_key=replace_key,
            skip_if_exists=True,
        )

    async def route_connection(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_records: List[MediationRecord],
    ) -> Optional[KeylistUpdate]:
        """Set up routing for a connection.

        This method will evaluate connection state and call the appropriate methods.
        """
        if conn_record.rfc23_state == ConnRecord.State.INVITATION.rfc23strict(
            ConnRecord.Role.RESPONDER
        ):
            return await self.route_connection_as_invitee(
                profile, conn_record, mediation_records
            )

        if conn_record.rfc23_state == ConnRecord.State.REQUEST.rfc23strict(
            ConnRecord.Role.REQUESTER
        ):
            return await self.route_connection_as_inviter(
                profile, conn_record, mediation_records
            )

        return None

    async def route_invitation(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_record: Optional[MediationRecord] = None,
    ) -> Optional[KeylistUpdate]:
        """Set up routing for receiving a response to an invitation."""
        await self.save_mediator_for_connection(profile, conn_record, mediation_record)

        if conn_record.invitation_key:
            LOGGER.debug("Routing invitation key %s", conn_record.invitation_key)
            return await self._route_for_key(
                profile,
                conn_record.invitation_key,
                mediation_record,
                skip_if_exists=True,
            )

        raise ValueError("Expected connection to have invitation_key")

    async def route_verkey(
        self,
        profile: Profile,
        verkey: str,
        mediation_record: Optional[MediationRecord] = None,
    ):
        """Establish routing for a public DID."""
        LOGGER.debug(
            "Routing verkey %s%s", verkey, " with mediation" if mediation_record else ""
        )
        return await self._route_for_key(
            profile, verkey, mediation_record, skip_if_exists=True
        )

    async def route_public_did(self, profile: Profile, verkey: str):
        """Establish routing for a public DID.

        [DEPRECATED] Establish routing for a public DID. Use route_verkey() instead.
        """
        return await self._route_for_key(profile, verkey, skip_if_exists=True)

    async def route_static(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_record: Optional[MediationRecord] = None,
    ) -> Optional[KeylistUpdate]:
        """Establish routing for a static connection."""
        LOGGER.debug("Routing static connection")
        my_info = await self.get_or_create_my_did(profile, conn_record)
        return await self._route_for_key(
            profile, my_info.verkey, mediation_record, skip_if_exists=True
        )

    async def save_mediator_for_connection(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_record: Optional[MediationRecord] = None,
        mediation_id: Optional[str] = None,
    ):
        """Save mediator info to connection metadata."""
        async with profile.session() as session:
            if mediation_id:
                mediation_record = await MediationRecord.retrieve_by_id(
                    session, mediation_id
                )

            if mediation_record:
                await conn_record.metadata_set(
                    session,
                    MediationManager.METADATA_KEY,
                    {MediationManager.METADATA_ID: mediation_record.mediation_id},
                )

    @abstractmethod
    async def routing_info(
        self,
        profile: Profile,
        mediation_record: Optional[MediationRecord] = None,
    ) -> RoutingInfo:
        """Retrieve routing keys."""

    async def connection_from_recipient_key(
        self, profile: Profile, recipient_key: str
    ) -> ConnRecord:
        """Retrieve connection for a recipient_key.

        The recipient key is expected to be a local key owned by this agent.
        """
        async with profile.session() as session:
            wallet = session.inject(BaseWallet)
            try:
                conn = await ConnRecord.retrieve_by_tag_filter(
                    session, {"invitation_key": normalize_from_did_key(recipient_key)}
                )
            except StorageNotFoundError:
                did_info = await wallet.get_local_did_for_verkey(
                    normalize_from_did_key(recipient_key)
                )
                conn = await ConnRecord.retrieve_by_did(session, my_did=did_info.did)

            return conn


class CoordinateMediationV1RouteManager(RouteManager):
    """Manage routes using Coordinate Mediation protocol."""

    async def _route_for_key(
        self,
        profile: Profile,
        recipient_key: str,
        mediation_record: Optional[MediationRecord] = None,
        *,
        skip_if_exists: bool = False,
        replace_key: Optional[str] = None,
    ) -> Optional[KeylistUpdate]:
        LOGGER.debug("Routing for key %s using coordinate mediation", recipient_key)
        if not mediation_record:
            return None

        if skip_if_exists:
            try:
                async with profile.session() as session:
                    await RouteRecord.retrieve_by_recipient_key(session, recipient_key)

                return None
            except StorageNotFoundError:
                pass

        # Keylist update is idempotent, skip_if_exists ignored
        mediation_mgr = MediationManager(profile)
        keylist_update = await mediation_mgr.add_key(recipient_key)
        if replace_key:
            keylist_update = await mediation_mgr.remove_key(replace_key, keylist_update)

        responder = profile.inject(BaseResponder)
        await responder.send(keylist_update, connection_id=mediation_record.connection_id)
        return keylist_update

    async def routing_info(
        self,
        profile: Profile,
        mediation_record: Optional[MediationRecord] = None,
    ) -> RoutingInfo:
        """Return routing info for mediator."""
        if mediation_record:
            return RoutingInfo(
                routing_keys=[
                    normalize_to_did_key(key).key_id
                    for key in mediation_record.routing_keys
                ],
                endpoint=mediation_record.endpoint,
            )

        return RoutingInfo.empty()



"""jsonld admin routes."""

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema
from marshmallow import INCLUDE, Schema, fields
from pydid.verification_method import Ed25519VerificationKey2018

from ...admin.decorators.auth import tenant_authentication
from ...admin.request_context import AdminRequestContext
from ...config.base import InjectionError
from ...resolver.base import ResolverError
from ...resolver.did_resolver import DIDResolver
from ...wallet.error import WalletError
from ..models.openapi import OpenAPISchema
from .credential import sign_credential, verify_credential
from .error import BaseJSONLDMessagingError

SUPPORTED_VERIFICATION_METHOD_TYPES = (Ed25519VerificationKey2018,)


class SignatureOptionsSchema(Schema):
    """Schema for LD signature options."""

    type = fields.Str(required=False)
    verificationMethod = fields.Str(required=True)
    proofPurpose = fields.Str(required=True)
    challenge = fields.Str(required=False)
    domain = fields.Str(required=False)


class DocSchema(OpenAPISchema):
    """Schema for LD doc to sign."""

    credential = fields.Dict(
        required=True, metadata={"description": "Credential to sign"}
    )
    options = fields.Nested(
        SignatureOptionsSchema,
        required=True,
        metadata={"description": "Signature options"},
    )


class SignRequestSchema(OpenAPISchema):
    """Request schema for signing a jsonld doc."""

    verkey = fields.Str(
        required=True, metadata={"description": "Verkey to use for signing"}
    )
    doc = fields.Nested(DocSchema, required=True)


class SignResponseSchema(OpenAPISchema):
    """Response schema for a signed jsonld doc."""

    signed_doc = fields.Dict(required=False, metadata={"description": "Signed document"})
    error = fields.Str(required=False, metadata={"description": "Error text"})


@docs(
    tags=["jsonld"],
    summary="Sign a JSON-LD structure and return it",
    deprecated=True,
)
@request_schema(SignRequestSchema())
@response_schema(SignResponseSchema(), 200, description="")
@tenant_authentication
async def sign(request: web.BaseRequest):
    """Request handler for signing a jsonld doc.

    Args:
        request: aiohttp request object

    """
    response = {}
    body = await request.json()
    doc = body.get("doc")
    try:
        context: AdminRequestContext = request["context"]
        async with context.session() as session:
            doc_with_proof = await sign_credential(
                session, doc.get("credential"), doc.get("options"), body.get("verkey")
            )
            response["signed_doc"] = doc_with_proof
    except BaseJSONLDMessagingError as err:
        response["error"] = str(err)
    except (WalletError, InjectionError):
        raise web.HTTPForbidden(reason="No wallet available")
    return web.json_response(response)


class SignedDocSchema(OpenAPISchema):
    """Verifiable doc schema."""

    class Meta:
        """Keep unknown values."""

        unknown = INCLUDE

    proof = fields.Nested(
        SignatureOptionsSchema,
        required=True,
        metadata={"unknown": INCLUDE, "description": "Linked data proof"},
    )


class VerifyRequestSchema(OpenAPISchema):
    """Request schema for signing a jsonld doc."""

    verkey = fields.Str(
        required=False, metadata={"description": "Verkey to use for doc verification"}
    )
    doc = fields.Nested(
        SignedDocSchema, required=True, metadata={"description": "Signed document"}
    )


class VerifyResponseSchema(OpenAPISchema):
    """Response schema for verification result."""

    valid = fields.Bool(required=True)
    error = fields.Str(required=False, metadata={"description": "Error text"})


@docs(
    tags=["jsonld"],
    summary="Verify a JSON-LD structure.",
    deprecated=True,
)
@request_schema(VerifyRequestSchema())
@response_schema(VerifyResponseSchema(), 200, description="")
@tenant_authentication
async def verify(request: web.BaseRequest):
    """Request handler for signing a jsonld doc.

    Args:
        request: aiohttp request object

    """
    response = {"valid": False}
    try:
        context: AdminRequestContext = request["context"]
        profile = context.profile
        body = await request.json()
        verkey = body.get("verkey")
        doc = body.get("doc")
        async with context.session() as session:
            if verkey is None:
                resolver = session.inject(DIDResolver)
                vmethod = await resolver.dereference(
                    profile,
                    doc["proof"]["verificationMethod"],
                )

                if not isinstance(vmethod, SUPPORTED_VERIFICATION_METHOD_TYPES):
                    raise web.HTTPBadRequest(
                        reason="{} is not supported".format(vmethod.type)
                    )
                verkey = vmethod.material

            valid = await verify_credential(session, doc, verkey)

        response["valid"] = valid
    except (BaseJSONLDMessagingError, ResolverError, ValueError) as error:
        response["error"] = str(error)
    except (WalletError, InjectionError):
        raise web.HTTPForbidden(reason="No wallet available")
    return web.json_response(response)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes([web.post("/jsonld/sign", sign), web.post("/jsonld/verify", verify)])


def post_process_routes(app: web.Application):
    """Amend swagger API."""
    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "jsonld",
            "description": "Sign and verify json-ld data",
            "externalDocs": {
                "description": "Specification",
                "url": "https://tools.ietf.org/html/rfc7515",
            },
        }
    )



from .credential import CredentialSchema as _CredentialSchema
from .credential import VerifiableCredential as _VerifiableCredential
from .credential import VerifiableCredentialSchema as _VerifiableCredentialSchema
from .linked_data_proof import LDProof as _LDProof
from .linked_data_proof import LinkedDataProofSchema as _LinkedDataProofSchema

__all__ = [
    "_VerifiableCredential",
    "_CredentialSchema",
    "_VerifiableCredentialSchema",
    "_LDProof",
    "_LinkedDataProofSchema",
]



"""Test RevokeHandler."""

from typing import Generator

import pytest
import pytest_asyncio

from ......core.event_bus import EventBus, MockEventBus
from ......core.profile import Profile
from ......messaging.request_context import RequestContext
from ......messaging.responder import BaseResponder, MockResponder
from ......utils.testing import create_test_profile
from ...messages.revoke import Revoke
from ..revoke_handler import RevokeHandler


@pytest.fixture
def event_bus():
    yield MockEventBus()


@pytest.fixture
def responder():
    yield MockResponder()


@pytest_asyncio.fixture
async def profile(event_bus):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(EventBus, event_bus)
    yield profile


@pytest.fixture
def message():
    yield Revoke(thread_id="mock_thread_id", comment="mock_comment")


@pytest.fixture
def context(profile: Profile, message: Revoke) -> Generator[RequestContext, None, None]:
    request_context = RequestContext(profile)
    request_context.message = message
    yield request_context


@pytest.mark.asyncio
async def test_handle(
    context: RequestContext, responder: BaseResponder, event_bus: MockEventBus
):
    context.connection_ready = True
    await RevokeHandler().handle(context, responder)
    assert event_bus.events
    [(_, received)] = event_bus.events
    assert received.topic == RevokeHandler.RECEIVED_TOPIC
    assert "thread_id" in received.payload
    assert "comment" in received.payload


@pytest.mark.asyncio
async def test_handle_monitor(
    context: RequestContext, responder: BaseResponder, event_bus: MockEventBus
):
    context.settings["revocation.monitor_notification"] = True
    context.connection_ready = True
    await RevokeHandler().handle(context, responder)
    [(_, webhook), (_, received)] = event_bus.events

    assert webhook.topic == RevokeHandler.WEBHOOK_TOPIC
    assert "thread_id" in webhook.payload
    assert "comment" in webhook.payload

    assert received.topic == RevokeHandler.RECEIVED_TOPIC
    assert "thread_id" in received.payload
    assert "comment" in received.payload



"""Presentation proposal message handler."""

from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import V20PresManager
from ..messages.pres_problem_report import ProblemReportReason
from ..messages.pres_proposal import V20PresProposal


class V20PresProposalHandler(BaseHandler):
    """Message handler class for presentation proposals."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for presentation proposals.

        Args:
            context: proposal context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug("V20PresProposalHandler called with context %s", context)
        assert isinstance(context.message, V20PresProposal)
        self._logger.debug(
            "Received v2.0 presentation proposal message: %s",
            context.message.serialize(as_string=True),
        )

        if not context.connection_record:
            raise HandlerException(
                "Connectionless not supported for presentation proposal"
            )
        # If connection is present it must be ready for use
        elif not context.connection_ready:
            raise HandlerException("Connection used for presentation proposal not ready")

        profile = context.profile
        pres_manager = V20PresManager(profile)
        pres_ex_record = await pres_manager.receive_pres_proposal(
            context.message, context.connection_record
        )  # mgr only creates, saves record: on exception, saving state err is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="V20PresProposalHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_respond_presentation_proposal is set, reply with proof req
        if context.settings.get("debug.auto_respond_presentation_proposal"):
            pres_request_message = None
            try:
                (
                    pres_ex_record,
                    pres_request_message,
                ) = await pres_manager.create_bound_request(
                    pres_ex_record=pres_ex_record,
                    comment=context.message.comment,
                )
                await responder.send_reply(pres_request_message)
            except (BaseModelError, LedgerError, StorageError) as err:
                self._logger.exception(err)
                if pres_ex_record:
                    async with profile.session() as session:
                        await pres_ex_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            pres_ex_record,
                            ProblemReportReason.ABANDONED.value,  # them: be vague
                        )
                    )

            trace_event(
                context.settings,
                pres_request_message,
                outcome="V20PresProposalHandler.handle.PRESENT",
                perf_counter=r_time,
            )



"""DIDComm v2 envelope handling via Askar backend."""

import json
from collections import OrderedDict
from typing import Mapping, Tuple, Union

from aries_askar import AskarError, Key, KeyAlg, Session, ecdh
from marshmallow import ValidationError

from ...utils.jwe import JweEnvelope, JweRecipient, b64url, from_b64url
from ...wallet.base import WalletError


class DidcommEnvelopeError(WalletError):
    """A base error class for DIDComm envelope wrapping and unwrapping operations."""


def ecdh_es_encrypt(to_verkeys: Mapping[str, Key], message: bytes) -> bytes:
    """Encode a message using DIDComm v2 anonymous encryption."""
    wrapper = JweEnvelope(with_flatten_recipients=False)

    alg_id = "ECDH-ES+A256KW"
    enc_id = "XC20P"
    enc_alg = KeyAlg.XC20P
    wrap_alg = KeyAlg.A256KW

    if not to_verkeys:
        raise DidcommEnvelopeError("No message recipients")

    try:
        cek = Key.generate(enc_alg)
    except AskarError:
        raise DidcommEnvelopeError("Error creating content encryption key")

    for kid, recip_key in to_verkeys.items():
        try:
            epk = Key.generate(recip_key.algorithm, ephemeral=True)
        except AskarError:
            raise DidcommEnvelopeError("Error creating ephemeral key")
        enc_key = ecdh.EcdhEs(alg_id, None, None).sender_wrap_key(
            wrap_alg, epk, recip_key, cek
        )
        wrapper.add_recipient(
            JweRecipient(
                encrypted_key=enc_key.ciphertext,
                header={"kid": kid, "epk": epk.get_jwk_public()},
            )
        )

    wrapper.set_protected(
        OrderedDict(
            [
                ("alg", alg_id),
                ("enc", enc_id),
            ]
        )
    )
    try:
        payload = cek.aead_encrypt(message, aad=wrapper.protected_bytes)
    except AskarError:
        raise DidcommEnvelopeError("Error encrypting message payload")
    wrapper.set_payload(payload.ciphertext, payload.nonce, payload.tag)

    return wrapper.to_json().encode("utf-8")


def ecdh_es_decrypt(
    wrapper: JweEnvelope,
    recip_kid: str,
    recip_key: Key,
) -> bytes:
    """Decode a message with DIDComm v2 anonymous encryption."""

    alg_id = wrapper.protected.get("alg")
    if alg_id in ("ECDH-ES+A128KW", "ECDH-ES+A256KW"):
        wrap_alg = alg_id[8:]
    else:
        raise DidcommEnvelopeError(f"Unsupported ECDH-ES algorithm: {alg_id}")

    recip = wrapper.get_recipient(recip_kid)
    if not recip:
        raise DidcommEnvelopeError(f"Recipient header not found: {recip_kid}")

    enc_alg = recip.header.get("enc")
    if enc_alg not in ("A128GCM", "A256GCM", "A128CBC-HS256", "A256CBC-HS512", "XC20P"):
        raise DidcommEnvelopeError(f"Unsupported ECDH-ES content encryption: {enc_alg}")

    try:
        epk = Key.from_jwk(recip.header.get("epk"))
    except AskarError:
        raise DidcommEnvelopeError("Error loading ephemeral key")

    apu = recip.header.get("apu")
    apv = recip.header.get("apv")

    try:
        cek = ecdh.EcdhEs(alg_id, apu, apv).receiver_unwrap_key(
            wrap_alg,
            enc_alg,
            epk,
            recip_key,
            recip.encrypted_key,
        )
    except AskarError:
        raise DidcommEnvelopeError("Error decrypting content encryption key")

    try:
        plaintext = cek.aead_decrypt(
            wrapper.ciphertext,
            nonce=wrapper.iv,
            tag=wrapper.tag,
            aad=wrapper.combined_aad,
        )
    except AskarError:
        raise DidcommEnvelopeError("Error decrypting message payload")

    return plaintext


def ecdh_1pu_encrypt(
    to_verkeys: Mapping[str, Key], sender_kid: str, sender_key: Key, message: bytes
) -> bytes:
    """Encode a message using DIDComm v2 authenticated encryption."""
    wrapper = JweEnvelope(with_flatten_recipients=False)

    alg_id = "ECDH-1PU+A256KW"
    enc_id = "A256CBC-HS512"
    enc_alg = KeyAlg.A256CBC_HS512
    wrap_alg = KeyAlg.A256KW
    agree_alg = sender_key.algorithm

    if not to_verkeys:
        raise DidcommEnvelopeError("No message recipients")

    try:
        cek = Key.generate(enc_alg)
    except AskarError:
        raise DidcommEnvelopeError("Error creating content encryption key")

    try:
        epk = Key.generate(agree_alg, ephemeral=True)
    except AskarError:
        raise DidcommEnvelopeError("Error creating ephemeral key")

    apu = b64url(sender_kid)
    apv = []
    for kid, recip_key in to_verkeys.items():
        if agree_alg:
            if agree_alg != recip_key.algorithm:
                raise DidcommEnvelopeError("Recipient key types must be consistent")
        else:
            agree_alg = recip_key.algorithm
        apv.append(kid)
    apv.sort()
    apv = b64url(".".join(apv))

    wrapper.set_protected(
        OrderedDict(
            [
                ("alg", alg_id),
                ("enc", enc_id),
                ("apu", apu),
                ("apv", apv),
                ("epk", json.loads(epk.get_jwk_public())),
                ("skid", sender_kid),
            ]
        )
    )
    try:
        payload = cek.aead_encrypt(message, aad=wrapper.protected_bytes)
    except AskarError:
        raise DidcommEnvelopeError("Error encrypting message payload")
    wrapper.set_payload(payload.ciphertext, payload.nonce, payload.tag)

    for kid, recip_key in to_verkeys.items():
        enc_key = ecdh.Ecdh1PU(alg_id, apu, apv).sender_wrap_key(
            wrap_alg, epk, sender_key, recip_key, cek, cc_tag=payload.tag
        )
        wrapper.add_recipient(
            JweRecipient(encrypted_key=enc_key.ciphertext, header={"kid": kid})
        )

    return wrapper.to_json().encode("utf-8")


def ecdh_1pu_decrypt(
    wrapper: JweEnvelope,
    recip_kid: str,
    recip_key: Key,
    sender_key: Key,
) -> Tuple[str, str, str]:
    """Decode a message with DIDComm v2 authenticated encryption."""

    alg_id = wrapper.protected.get("alg")
    if alg_id in ("ECDH-1PU+A128KW", "ECDH-1PU+A256KW"):
        wrap_alg = alg_id[9:]
    else:
        raise DidcommEnvelopeError(f"Unsupported ECDH-1PU algorithm: {alg_id}")

    enc_alg = wrapper.protected.get("enc")
    if enc_alg not in ("A128CBC-HS256", "A256CBC-HS512"):
        raise DidcommEnvelopeError(f"Unsupported ECDH-1PU content encryption: {enc_alg}")

    recip = wrapper.get_recipient(recip_kid)
    if not recip:
        raise DidcommEnvelopeError(f"Recipient header not found: {recip_kid}")

    try:
        epk = Key.from_jwk(wrapper.protected.get("epk"))
    except AskarError:
        raise DidcommEnvelopeError("Error loading ephemeral key")

    apu = wrapper.protected.get("apu")
    apv = wrapper.protected.get("apv")

    try:
        cek = ecdh.Ecdh1PU(alg_id, apu, apv).receiver_unwrap_key(
            wrap_alg,
            enc_alg,
            epk,
            sender_key,
            recip_key,
            recip.encrypted_key,
            cc_tag=wrapper.tag,
        )
    except AskarError:
        raise DidcommEnvelopeError("Error decrypting content encryption key")

    try:
        plaintext = cek.aead_decrypt(
            wrapper.ciphertext,
            nonce=wrapper.iv,
            tag=wrapper.tag,
            aad=wrapper.combined_aad,
        )
    except AskarError:
        raise DidcommEnvelopeError("Error decrypting message payload")

    return plaintext


async def unpack_message(
    session: Session, enc_message: Union[bytes, str]
) -> Tuple[str, str, str]:
    """Decode a message using DIDComm v2 encryption."""
    try:
        wrapper = JweEnvelope.from_json(enc_message)
    except ValidationError:
        raise DidcommEnvelopeError("Invalid packed message")

    alg = wrapper.protected.get("alg")
    method = next((m for m in ("ECDH-1PU", "ECDH-ES") if m in alg), None)
    if not method:
        raise DidcommEnvelopeError(f"Unsupported DIDComm encryption algorithm: {alg}")

    sender_key = None
    sender_kid = None
    recip_key = None
    recip_kid = None
    for kid in wrapper.recipient_key_ids:
        recip_key_entry = next(
            iter(await session.fetch_all_keys(tag_filter={"kid": kid})), None
        )
        if recip_key_entry:
            recip_kid = kid
            recip_key = recip_key_entry.key
            break

    if not recip_key:
        raise DidcommEnvelopeError("No recognized recipient key")

    if method == "ECDH-1PU":
        sender_kid_apu = None
        apu = wrapper.protected.get("apu")
        if apu:
            try:
                sender_kid_apu = from_b64url(apu).decode("utf-8")
            except (UnicodeDecodeError, ValidationError):
                raise DidcommEnvelopeError("Invalid apu value")
        sender_kid = wrapper.protected.get("skid") or sender_kid_apu
        if sender_kid_apu and sender_kid != sender_kid_apu:
            raise DidcommEnvelopeError("Mismatch between skid and apu")
        if not sender_kid:
            raise DidcommEnvelopeError("Sender key ID not provided")
        # FIXME - validate apv if present?
        # FIXME - will need to insert proper sender key resolution method here
        # instead of looking in the wallet
        sender_key_entry = next(
            iter(await session.fetch_all_keys(tag_filter={"kid": sender_kid})), None
        )
        if not sender_key_entry:
            raise DidcommEnvelopeError("Sender public key not found")
        sender_key = sender_key_entry.key
        plaintext = ecdh_1pu_decrypt(wrapper, recip_kid, recip_key, sender_key)
    else:
        plaintext = ecdh_es_decrypt(wrapper, recip_kid, recip_key)

    return plaintext, recip_kid, sender_kid



"""Multiledger related utility methods."""

import logging
from collections import OrderedDict

from ..config.settings import BaseSettings
from ..core.error import ProfileError

LOGGER = logging.getLogger(__name__)


def get_write_ledger_config_for_profile(settings: BaseSettings) -> dict:
    """Return initial/default write ledger config on profile creation."""
    write_ledger_config = None
    prod_write_ledger_pool = OrderedDict()
    non_prod_write_ledger_pool = OrderedDict()

    LOGGER.debug("Getting write ledger config for profile")
    for ledger_config in settings.get("ledger.ledger_config_list"):
        is_production = ledger_config.get("is_production")
        is_write = ledger_config.get("is_write")
        is_read_only = ledger_config.get("read_only")
        ledger_id = ledger_config.get("id") or ledger_config.get("pool_name")

        if is_production and (is_write or is_read_only):
            prod_write_ledger_pool[ledger_id] = ledger_config
        elif not is_production and (is_write or is_read_only):
            non_prod_write_ledger_pool[ledger_id] = ledger_config
        else:
            LOGGER.warning(
                "Ledger config %s is not a write ledger nor a read-only ledger",
                ledger_id,
            )

    write_ledger = settings.get("ledger.write_ledger")
    if write_ledger:
        if write_ledger in prod_write_ledger_pool:
            write_ledger_config = prod_write_ledger_pool.get(write_ledger)
        elif write_ledger in non_prod_write_ledger_pool:
            write_ledger_config = non_prod_write_ledger_pool.get(write_ledger)
        else:
            error_message = (
                "ledger.write_ledger in profile settings does not correspond to a "
                "write configurable ledger provided with --genesis-transactions-list"
            )
            LOGGER.error(error_message)
            raise ProfileError(error_message)
    else:
        if len(prod_write_ledger_pool) >= 1:
            LOGGER.debug("Using first production write ledger")
            write_ledger_config = (list(prod_write_ledger_pool.values()))[0]
        elif len(non_prod_write_ledger_pool) >= 1:
            LOGGER.debug("Using first non-production write ledger")
            write_ledger_config = (list(non_prod_write_ledger_pool.values()))[0]
        else:
            LOGGER.error("No write ledger configuration found in ledger_config_list")
            raise ProfileError(
                "No write ledger configuration found in ledger_config_list which "
                "was provided with --genesis-transactions-list"
            )

    return write_ledger_config



from unittest import IsolatedAsyncioTestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from ......tests import mock
from ......utils.testing import create_test_profile
from ...message_types import ATTACHMENT_FORMAT, PRES_20_PROPOSAL
from ...messages.pres_format import V20PresFormat
from ...messages.pres_proposal import V20PresProposal
from .. import pres_exchange as test_module
from ..pres_exchange import V20PresExRecord

S_ID = "NcYxiDXkpYi6ov5FcYDi1e:2:vidya:1.0"
CD_ID = f"NcYxiDXkpYi6ov5FcYDi1e:3:CL:{S_ID}:tag1"
INDY_PROOF_REQ = {
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {
        "0_player_uuid": {
            "name": "player",
            "restrictions": [
                {
                    "cred_def_id": f"{CD_ID}",
                    "attr::player::value": "Richie Knucklez",
                }
            ],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        },
        "0_screencapture_uuid": {
            "name": "screenCapture",
            "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        },
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        }
    },
}


class BasexRecordImpl(BaseExchangeRecord):
    class Meta:
        schema_class = "BasexRecordImplSchema"

    RECORD_TYPE = "record"


class BasexRecordImplSchema(BaseExchangeSchema):
    class Meta:
        model_class = BasexRecordImpl


class TestRecord(IsolatedAsyncioTestCase):
    async def test_record(self):
        pres_proposal = V20PresProposal(
            comment="Hello World",
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.INDY.api
                    ],
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            verified="false",
            auto_present=True,
            error_msg="error",
            auto_remove=True,
        )
        record.pres_proposal = pres_proposal  # cover setter

        assert record.pres_ex_id == "pxid"

        assert record.record_value == {
            "connection_id": "conn_id",
            "initiator": "init",
            "role": "role",
            "state": "state",
            "pres_proposal": pres_proposal.serialize(),
            "verified": "false",
            "verified_msgs": None,
            "auto_present": True,
            "auto_verify": False,
            "error_msg": "error",
            "trace": False,
            "auto_remove": True,
        }

        bx_record = BasexRecordImpl()
        assert record != bx_record

    async def test_save_error_state(self):
        self.profile = await create_test_profile()
        record = V20PresExRecord(state=None)
        assert record._last_state is None
        async with self.profile.session() as session:
            await record.save_error_state(session)  # cover short circuit

            record.state = V20PresExRecord.STATE_PROPOSAL_RECEIVED
            await record.save(session)

            with (
                mock.patch.object(record, "save", mock.CoroutineMock()) as mock_save,
                mock.patch.object(
                    test_module.LOGGER, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                mock_save.side_effect = test_module.StorageError()
                await record.save_error_state(session, reason="testing")
                mock_log_exc.assert_called_once()
